;*****************************************************************************
;*****************************************************************************
;
;	ISR.ASM - Interrupt Service Reflector
;
;	Copyright (c) 1991
;
;	Rich Bono (NM1D)
;	7 Redfield Circle
;	Derry, New Hampshire, 03038
;
;	All Rights Reserved
;
;
;	This is designed to retain a specified interrupt vector while using
;	a debugger under MS-DOS that swaps interrupt vectors during use.
;
;	This uses interrupt 69h for a command interface.
;	The specified vector is also used.
;
;	Requires:
;
;	  DOS 3.0 or later
;
;	  Interrupt vector 69h (105d) must not be in use by any other
;	  application or device driver.
;
;	  version 1.00 uses 480 bytes when installed
;
;    usage:
;
;    ISR [-/][vV]X [-/][rR]
;
;	-Vx  install into vector number 'x' (where 'x' is a hex number 8-F)
;	-R - Remove
;
;	defaults to vector 0Ch (COM1)
;	to remove a vector other than 0Ch, you must specify the vector number
;       and the remove switch.
;
;    i.e.: To install ISR into vector 0Bh:
;
;         ISR /vB
;
;          to remove ISR from vector 0Bh:
;
;         ISR /vB /r
;
;  CMDINT usage:
;
;	Note: CMDINT should only be invoked if the resident service is
;	      installed.
;	      If possible use the provided support routine _ISReflect,
;	      see the file ISREFLEC.ASM
;
;   Input:
;	AX must be 06996h or nothing will be done
;
;	DS:DX should be the segement:offset of your ISR that you would like
;             to be active during debug sessions.
;
;	DS:DX should be 0:0 to turn off function.
;
;   Output: AX == 09669h
;
;	For example:
;
;		;
;		; Perform the following just AFTER installing your ISR vector
;		; to turn on reflection of your ISR
;		;
;		Push ds
;		Mov dx, seg YourIsr		;Load segment of YourIsr -> DS
;		Mov ds, dx			;
;		Mov dx, offset YourIsr		;Load offset of YourIsr -> DX
;		Mov ax, 06996h			;Special cmd FLAG
;		Int 069h			; AX is 09669h on return
;		Pop ds
;
;		.
;		.
;		.
;
;		;
;		; Just BEFORE removing your ISR vector, do the following
;		; to indicate that reflection should be turned off
;		;
;		Push ds
;		Mov dx, 0			;Set segment and offset to 0:0
;		Mov ds, dx			;
;		Mov ax, 09669h			;Special cmd FLAG
;		Int 069h			; AX is 09669h on return
;		Pop ds
;
;
;
;	Revisions:
;
;	1.00 21-Jan-91	First version
;
;
;*****************************************************************************

.MODEL	small

include isr.inc

;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------
StdOut MACRO txt, lngth		;Output text to standard output
	mov	ah, 040h		;Write file or device
	mov	bx, 1			;STDOUT
	mov	cx, lngth		;Length of text to output
	mov	dx, offset txt		;Text to output
	int	021h
       ENDM

StdErr MACRO txt, lngth		;Output text to standard error output
	mov	ah, 040h		;Write file or device
	mov	bx, 2			;STDERR
	mov	cx, lngth		;Length of text to output
	mov	dx, offset txt		;Text to output
	int	021h
       ENDM
;-----------------------------------------------------------------------------
;-----------------------------------------------------------------------------

;*****************************************************************************

.CODE
		;
		;This is here to ID the binary, don't change location or
		; content! (must match ISREFLEC.ASM)
		;
Banner	label byte
	BANNERSTRING		;Generate the banner (defined in ISR.INC)
BANNERLEN equ $ - Banner	; length of the Banner string
	db	EOF

;*****************************************************************************

OrigOff	dw	0000h		; Original address stored here
OrigSeg	dw	0000h

OrigCmdOff dw	0000h		;Original CMDINT vector is stored here
OrigCmdSeg dw	0000h

ORG_PSP	dw	0000h		;Original PSP segment


PreferredVector	label dword	;Installed vector
NewOff	dw	0000h
NewSeg	dw	0000h

SavedVector label dword		; vector
SavOff	dw	0000h		; int vector offset
SavSeg	dw	0000h		; int vector segment

IntVect		db	THISINT	;Interrupt vector to reflect

;*****************************************************************************

MyISR	proc	far			;Do this for each interrupt

	pushf				;Save current flags

	push	ax			;Save used registers
	mov	ax, cs:NewOff
	or	ax, cs:NewSeg		;Check for 0:0 special case
	pop	ax

	jz	short NoReflect		; No vector to reflect, continue
					; with original . . .
	;
	; If here, then there does seem to be a preferred vector,
	; please use this instead!
	;
	popf				;Restore flags
	jmp	cs:[PreferredVector]	; Continue with the desired vector

NoReflect:
	popf				;Restore flags
	jmp	cs:[SavedVector]	; Continue with the original vector

MyISR	endp


;*****************************************************************************
;*****************************************************************************
;
; SetPreferred - This is the ISR command interface.  
;
;	This will simply save the contents of DS:DX into PreferredVector.
;
;   Input:
;	AX must be 06996h or nothing will be done
;
;	DS:DX should be the segement:offset of your ISR that you would like
;             to be active during debug sessions.
;
;	DS:DX should be 0:0 to turn off function.
;
;   Output: AX == 09669h
;

SetPreferred proc far
	;
	; Set vector according to DS:DX
	;
	cmp	ax, S_FLAG		;Sanity check, is AX==S_FLAG?
	jne	short NotCorrect	; No, so don't do anything

	push	dx			;Save used register
	mov	cs:NewOff, dx		; Store desired offset
	mov	dx, ds			;
	mov	cs:NewSeg, dx		; Store desired segment
	pop	dx

	mov	ax, E_FLAG		;Return special flag to show active

NotCorrect:

	iret				;All done!
SetPreferred endp

;*****************************************************************************
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************
	  ;
EndOfTSR: ; ***This is the LAST BYTE of the TSR that will remain resident!!!**
	  ;
	  ; EVERYTHING past here will will be TRASHED after installation
	  ;
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************
;*****************************************************************************


;---------------------------------------------------------------------------
;
; Install (tsr) - This will check the environment, and if all looks ok,
;                 install the replacement vector MyISR to be the
;                 Interrupt Service Routine.  CMDINT is also installed.
;
;
; This is exec'ed from MS-DOS.  It performs the following functions:
;
;	1) Checks for MS-DOS 3.0 or newer
;
;	2) Scan the command line for startup switches 
;	    
;	3) Checks to see that this has not already been installed (checks the
;	    code that the ISR vector points to).  !!!! Note: this will not
;	    properly identify installation if something else hooks the same
;	    interrupt vectors after installation!!!!!!!!
;
;	4) Install (release ram from the Environment)
;
;	5) Exit (TSR) return code of 0
;
;     If any errors are detected a value of 1 is returned
;
;---------------------------------------------------------------------------

Install proc
	;
	; When entered here CS:IP -> Install
	;                   SS:SP -> Stack
	;                   DS,ES -> PSP


	mov	ax, es		;Get segment of PSP
	mov	cs:ORG_PSP, ax	; save into our CODE SEGMENT

	mov	dx, cs		;This banner is in the code segment
	mov	ds, dx
	assume	ds:_TEXT
	StdOut Banner, BANNERLEN

	mov	ax, @data	; Get our data segment
	mov	ds, ax
	assume  ds:@data

;***************************************************************************
;
; Check to see that DOS is 3.x or newer
;
;***************************************************************************
	mov	ah, 30h		;DOS get version
	mov	al, 0h
	int	DOSINT
	cmp	al, 3		;Version 3.x?
	jae	DosOK		;Yes . . .

	StdErr	BadDOSMSG, BadDOSLen
	jmp	ErrExit

DosOK:

;***************************************************************************
;
; Scan the command line and check for the remove switch
;
;	Command line is stored at PSP:81h
;                                (PSP:80h is length of the command line)
;
;	supports:
;
;	switches [/-]
;                 [Rr]	- remove
;                 [vV][x]  vector number (x) from 8-Fh
;
;***************************************************************************
ScanCmdLine:
	push	ds		;Save current data segment
	mov	ax, ds
	mov	es, ax		; ES -> current data segment

	assume	es:@data

	mov	ax, ORG_PSP	;Get address of our PSP
	mov	ds, ax		; DS->PSP

	assume	ds:_TEXT

	xor	cx, cx		;Zero CX
	mov	cl, ds:[80h]	; Get length into CL
	mov	si, 081h	; Point DI at the command line
	cld			; direction UP

CmdScan: jcxz	CmdDone		;No command line to scan
	lodsb			; Get command line byte into AL

	cmp	al, ' '		;If any whitespace, ignore
	jz	CmdLoop
	cmp	al, 09h		;Tab
	jz	CmdLoop
	cmp	al, 0ah		;LF
	jz	CmdLoop
	cmp	al, 0dh		;CR
	jz	CmdLoop

	cmp	al, '-'		;Switch character
	jz	HaveSwitch
	cmp	al, '/'
	jz	HaveSwitch

	jmp	short CmdErr	;Not a valid command . . .

CmdLoop: loop	CmdScan		;Get next byte if any

	jmp	CmdDone 	;Done with commands (was only whitespace) . . 

;***************************************************************************
;
; If here, didn't find proper command line, give help
;
;***************************************************************************
CmdErr:	pop	ds		;Restore DS

	assume	ds:@data

	StdErr	CmdHelpMsg, CmdHelpLen
	jmp	ErrExit

;***************************************************************************
HaveSwitch:	;Come here if found a switch character
	dec	cx		;Count down switch character
	jcxz	CmdErr		; No options... error

	lodsb			;Read option

	cmp	al, 'r'		;remove switch?
	jz	SetRemove	; Yes . . .
	cmp	al, 'R'
	jz	SetRemove

	cmp	al, 'v'		;Vector switch?
	jz	SetVector	; Yes . . .
	cmp	al, 'V'
	jz	SetVector

	jmp	CmdErr		;Bad command line, exit . . .

;***************************************************************************
SetRemove:			;OK, got the remove switch
	mov	es:remove, 1	; Set flag to force remove
	jmp	short CmdLoop	; Handle rest of command line

SetVector:			;Got the set vector switch
	dec	cx		; Count down this character (the 'V')
	jcxz	CmdErr		;  If no more characters, then error . . .

	lodsb			;Read the Vector number
	cmp	al, 'A'
	jb	short notLetter
	and	al, NOT 'a'-'A' ; Be sure character is UPPER case
notLetter:
	cmp	al, '0'
	jb	CmdErr		;Less than '0', error
	cmp	al, 'F'
	ja	CmdErr		;Higher than 'F', error

	; Convert to Hex character
	sub	al, '0'		;remove ASCII offset
	cmp	al, 9		;Larger than 9?
	jbe	cnvt_done	;No . . .

	sub	al, 'A'-'9'-1	;remove offset
	cmp	al, 0fh		;larger than F?
	ja	CmdErr		; yes, error. . .

cnvt_done:
	;
	; Should now have a value, check for valid range
	;
	cmp	al, 08h		;Can't be less than 8
	jb	CmdErr		; Illegal
	cmp	al, 0Fh		;Can't be higher than 0Fh
	ja	CmdErr

	;
	; Ok, have a legal vector number, install it
	;
	mov	cs:IntVect, al	;Store vector number

	jmp	short CmdLoop	; Finish with the rest of the command line

CmdDone: pop	ds		;Restore DS

	assume	ds:@data

OkToInstall:

;***************************************************************************
; Ok to install, start process, get address of current ISR code
;***************************************************************************
CheckIntVect:
	; Get address of original ISR
	mov	ah, GIV			;DOS GIV
	mov	al, cs:IntVect		; ISR interrupt to hook
	int	DOSINT			; ES:BX now has interrupt vector

;***************************************************************************
; ES:BX now contains the original pointer.  Check to see if this points
;  to code that looks just like MyISR.
;***************************************************************************

	assume	es:NOTHING

	mov	cs:OrigOff, bx
	mov	cs:OrigSeg, es		; Save vector

	push	ds
	mov	di, bx			;Get offset into DI for compare
	mov	bx, cs
	mov	ds, bx			;Get current CS into DS

	assume	ds:_TEXT

	;
	; compare: DS:SI is pointing at this code
	;          ES:DI is pointing at installed code
	; 

	mov	si, offset MyISR	; Get offset of code to check
	mov	cx, offset EndOfTSR -  offset MyISR
	repe cmpsb			; Compare code
	pop	ds

	assume	ds:@data

	jne	short checkCmdInt
	jmp	IsInstalled	; Vector is still pointing at our code  . . .

checkCmdInt:
;***************************************************************************
;  If here, then code was NOT found at IntVect.  It may just be that
;  IntVect was taken over by another application, to check this
;  condition, check CMDTINT as a double check.
;
; Note: This is NOT failsafe, if BOTH vectors are taken over, all hope is
;       lost!
;
;***************************************************************************
; Before installing check to be sure that this is not already installed
;  (allow this to be installed only once!!!)
;***************************************************************************
	;
	; Check to see if the command vector is pointing at our code
	;
	; Get address of original ISR
	mov	ah, GIV			;DOS GIV
	mov	al, CMDINT		; ISR vector to check
	int	DOSINT			; ES:BX now has interrupt vector

	assume	es:NOTHING

	push	ds
	mov	di, bx			;Get offset into DI for compare
	mov	bx, cs
	mov	ds, bx			;Get current CS into DS

	assume	ds:_TEXT

	;
	; compare: DS:SI is pointing at this code
	;          ES:DI is pointing at installed code
	; 
	mov	si, offset SetPreferred	; Get offset of code to check
	mov	cx, offset EndOfTSR -  offset SetPreferred 
	repe cmpsb			; Compare code
	pop	ds

	assume	ds:@data

	jne	NotInstalled	; Vector is NOT pointing at our code  . . .


	;
	; If get here, then CMTINT is installed, but IntVect is not pointing
	; at our code, so this can't be removed.
	;


	test	remove, 1	;Is remove switch set?
	jnz	NotRemovable	;Yes . . .

	jmp	Abort		; No . . . Tell user already installed . . .

NotRemovable:
	;
	; Inform user that ISR is installed and CANNOT be removed
	; because another program has stolen the IntVect vector
	;

	StdErr	CantRemoveMsg, CantRemoveLen
	jmp	ErrExit

;***************************************************************************
;
; Code is not installed,
;  Check that the remove switch has not been specified
;
;***************************************************************************
NotInstalled:
	test	remove, 1	;Is remove switch set?
	jz	InstallIt	; No . . .

	;
	;Remove flag has been set, but our code is not installed,
	; inform user of error
	;

	StdErr	NotInstMsg, NotInstLen
	jmp	ErrExit

;***************************************************************************
; InstallIt - Code is not resident and needs to be installed
;***************************************************************************
InstallIt:

	; Inform user that installation is happening

	StdOut	OKMSG, OKLen

	mov	al, cs:IntVect		;Get vector number
	add	al, '0'			;Convert to ASCII
	cmp	al, '9'
	jbe	noGap
	add	al, 'A'-'9'-1		;account for hex gap
noGap:
	mov	ascVect, al		;Store vector number in message
	StdOut	VectMsg, VectLen	;Show vector that is installed

;***************************************************************************
; Get rid of unused environment (release memory to DOS)
;***************************************************************************
 	mov	ax, ORG_PSP		;Get original PSP segment
 	mov	es, ax			; into ES

 	assume	es:NOTHING

 	mov	es, es:[ENVIRON]	; Get segment of envirionment
 	mov	ah, FREEMEM		;DOS Free memory block
	int	DOSINT

;***************************************************************************
; Take over the IntVect vector
;***************************************************************************
	mov	al, cs:IntVect		;Going to hook into interrupt function
	mov	ah, GIV			; Get current int vector
	int	DOSINT			; ES:BX now contain vector

	assume	es:NOTHING

	mov	cs:SavOff, bx		;Store offset
	mov	cs:SavSeg, es		;Store segment

	mov	ah, SIV			;DOS SIV
	mov	al, cs:IntVect		; Interrupt vector to set
	mov	dx, SEG MyISR		; Get current CS into DS
	mov	ds, dx

	assume	ds:_TEXT

	mov	dx, offset MyISR	; Get offset into DX
	int	DOSINT

	;
	; Get current CMDINT vector
	;
	mov	al, CMDINT		;Vector to save
	mov	ah, GIV			; Get current int vector
	int	DOSINT			; ES:BX now contains vector
	mov	cs:OrigCmdOff, bx	;Store offset
	mov	cs:OrigCmdSeg, es	;Store segment

	; And install command vector
	mov	dx, seg SetPreferred	;Install SetPreferred address
	mov	ds, dx			; Into desired vector
	mov	dx, offset SetPreferred
	mov	al, CMDINT		;Command interrupt
	mov	ah, SIV
	int	DOSINT

;***************************************************************************
; Terminate, but leave code (up to EndOfTSR) resident
;***************************************************************************

	mov   dx,offset EndOfTSR ; Last byte of program to keep resident
	add   dx, 100h		; Don't forget 100 offset!
	mov   cl,4		;  Convert into paragraphs
	shr   dx,cl		;  (divide by 16)
	inc   dx		;  Plus one for round/off
				;DX specifies number of paragraphs to remain
				; resident
	mov   al, 0		;Return code of 0 (OK)
	mov   ah, TSR		;DOS TSR
	int   DOSINT		; Bye . . .

;***************************************************************************
; Come here if code is found to be installed
;***************************************************************************
IsInstalled:
	assume	ds:@data

	test	remove, 1	;Check if remove switch has been specified
	jz	Abort		; No, so inform user already installed

;***************************************************************************
; Remove the currently installed TSR
;***************************************************************************
	;
	; Get segment of the installed ISR into ES
	mov	ah, GIV		;DOS Get Interrupt Vector
	mov	al, cs:IntVect	; interrupt vector
	int	DOSINT		; ES:BX now has interrupt vector

	assume	es:NOTHING

;***************************************************************************
; Check to be sure that an application is not depending on our code
;***************************************************************************

	mov	ax, es:NewOff	;Read offset
	or	ax, es:NewSeg	; and segment
	jz	NotInUse	; If zero then not in use, ok to remove

	StdErr	InUseMsg, InUseLen ; Inform user why this can't be removed
	jmp	ErrExit		; exit

NotInUse:

;***************************************************************************
; Restore orginal ISR address into interrupt vector
;***************************************************************************
	mov	ah, SIV		;DOS Set Interrupt Vector
	mov	al, cs:IntVect	; interrupt vector to restore
	mov	dx, ES:OrigSeg	; Get original segment into DS

	push	ds		;Save our DS
	mov	ds, dx

	assume	ds:NOTHING

	mov	dx, ES:OrigOff	; Get original offset into DX
	int	DOSINT		;

	;
	; Restore original CMDINT vector
	;

	mov	ah, SIV			;DOS Set Interrupt Vector
	mov	al, CMDINT		; interrupt vector to set
	mov	dx, ES:OrigCmdSeg	; Get original segment into DS
	mov	ds, dx
	mov	dx, ES:OrigCmdOff	; and original offset into DX
	int	DOSINT			;Restore vector

	pop	ds		;Restore our DS

	assume	ds:@data

;***************************************************************************
; Give RAM back to DOS
;***************************************************************************

	; ES points to RAM block to free
	mov	es, es:ORG_PSP	;Get address of original PSP into ES
 	mov	ah, FREEMEM	;DOS Free memory block
 	int	DOSINT

	jc	removeErr	;Error returning memory!

	StdOut	RemovedMsg, RemovedLen
	jmp	Exit

;***************************************************************************
removeErr:	; Inform user that an error occured trying to free RAM
	StdErr	RemoveRamMsg, RemoveRamLen
	jmp	ErrExit

;***************************************************************************
; Jmp here to abort with message informing user that already installed
Abort:
	StdErr	AbortMSG, AbortLen

;***************************************************************************
; jmp here to exit with error code (no code resident)

ErrExit:
	mov	ah, EXIT	;DOS terminate
	mov	al, 1		; return code of 1 (problem)
	int	DOSINT		;Bye . . .

;***************************************************************************
Exit:	;Exit without error (not Resident)
	mov	al, 0		;Return code of 0 (no errors)
	mov	ah, EXIT	;DOS terminate
	int	DOSINT

Install endp

;***************************************************************************
; Local data - NOTE!!!! This is NOT resident data, it will disappear when
;              the TSR is installed!!!
;***************************************************************************
.DATA

remove	dw	0		;remove flag, when not zero, means to remove
				; TSR (set by /r switch)

AbortMSG	db       "ISR is already installed.",CR,LF
AbortLen	equ $-AbortMSG

OKMSG		db       "ISR installed.",CR,LF
OKLen		equ $-OKMSG

BadDOSMSG	db       "ISR requires MS-DOS version 3.0 or later."
		db       CR,LF
BadDOSLen	equ $-BadDOSMSG

CmdHelpMsg	db	 "ISR bad command line!", CR, LF
		db	 "  usage:", CR,LF
		db	 "   to install: ISR        (default vector 0Ch "
		db	 "- COM1)", CR, LF
		db	 "               ISR /vX    (where X is a vector from "
		db	 "8-Fh IRQ0-IRQ7)"
		db	 CR, LF,LF
		db	 "   to remove:  ISR /r     (default vector only)"
		db	 CR, LF
		db	 "               ISR /vX /r (remove non-default "
		db	 "vector)"
		db	CR, LF
CmdHelpLen	equ $-CmdHelpMsg

NotInstMsg	db	 "ISR is not installed, can't remove."
		db	CR,LF
NotInstLen	equ $-NotInstMsg

RemovedMsg	db	 "ISR has been removed.", CR, LF
RemovedLen	equ $-RemovedMsg

RemoveRamMsg	db	 "ISR has been removed but could not release "
		db	"memory!", CR, LF
RemoveRamLen	equ $-RemoveRamMsg

InUseMsg	db	 "ISR can't be removed while another application"
		db	 "is using it.", CR, LF
		db       "  Terminate the other application first.", CR, LF
InUseLen	equ $-InUseMsg

CantRemoveMsg	db	 "ISR can't be removed because another application "
		db       "has taken over",CR,LF,"         the target vector!"
		db	 CR,LF
CantRemoveLen	equ $-CantRemoveMsg

VectMsg		db	 " Interrupt Vector number : 0"
ascVect		db	'C',"h", CR, LF
VectLen		equ $-VectMsg
;***************************************************************************
.STACK 2048	;Stack of 2k (not installed)
;***************************************************************************

	END Install

; end of ISR.ASM
